import collections.abc
import typing
import datetime

import zodchy
from dataclasses import dataclass, fields

from .types import TaskId, MessageId, Dispatcher
from .enums import TaskStatus


class Message(typing.TypedDict):
    id: MessageId
    name: str
    body: dict
    headers: dict | None = None


@dataclass(frozen=True)
class Query(zodchy.codex.cqea.Query):
    def __iter__(
        self,
    ) -> collections.abc.Iterable[tuple[str, zodchy.codex.query.ClauseBit]]:
        for field in fields(self):
            value = getattr(self, field.name)
            if value is not zodchy.types.Empty:
                yield field.name, value


@dataclass(frozen=True, slots=True, kw_only=True)
class OutboxMessageCreated(zodchy.codex.cqea.Event):
    id: MessageId
    name: str
    body: dict
    headers: dict | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class OutboxTaskCreated(zodchy.codex.cqea.Event):
    id: TaskId
    message_id: MessageId
    dispatcher: Dispatcher
    scheduled_at: datetime.datetime
    settings: dict | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class GetOutboxTasks(Query):
    id: zodchy.operators.FilterBit[TaskId] | zodchy.types.Empty = zodchy.types.Empty
    dispatcher_id: zodchy.operators.FilterBit[str] | zodchy.types.Empty = (
        zodchy.types.Empty
    )
    status: zodchy.operators.FilterBit[str] | zodchy.types.Empty = zodchy.types.Empty
    scheduled_at: zodchy.operators.FilterBit[datetime.datetime] | zodchy.types.Empty = (
        zodchy.types.Empty
    )
    limit: zodchy.operators.SliceBit | zodchy.types.Empty = zodchy.types.Empty


@dataclass(frozen=True, slots=True, kw_only=True)
class OutboxTaskUpdated(zodchy.codex.cqea.Event):
    id: TaskId
    status: TaskStatus
    scheduled_at: datetime.datetime


@dataclass(frozen=True, slots=True, kw_only=True)
class OutboxTaskForProcessingReceived(zodchy.codex.cqea.Event):
    id: TaskId
    scheduled_at: datetime.datetime
    message: Message
    dispatcher_id: str
    settings: dict | None = None



import sqlalchemy

from .generic import TransactionWriter
from ....contracts.storage import MessageCreationModel
from ....contracts.types import MessageId


class MessageCreationWriter(TransactionWriter):
    async def __call__(self, message: MessageCreationModel) -> MessageId:
        await self._transaction.execute(
            sqlalchemy.insert(self._schema.messages).values(
                **{k: v for k, v in message.items() if v is not None},
            )
        )
        return message["id"]



from .tasks import TasksCreationWriter, TasksUpdatingWriter, TasksStatusUpdatingWriter
from .messages import MessageCreationWriter

__all__ = [
    "TasksCreationWriter",
    "TasksUpdatingWriter",
    "TasksStatusUpdatingWriter",
    "MessageCreationWriter",
]



import typing
from zodchy.codex.transport import DispatcherContract, CommunicationMessage
from .contracts.messages import GetOutboxTasks, TaskStatus
from .contracts.types import TaskId, MessageId
from .contracts.storage import (
    TasksForStatusReaderContract,
    TasksStatusUpdatingWriterContract,
)
from .actors.processing import tasks_ready_for_processing_reader
from .contracts.messages import OutboxTaskForProcessingReceived


class ZodchyOutboxDispatcherRegistry:
    def __init__(self):
        self._dispatchers = {}

    def register(
        self,
        id: str,
        dispatcher: DispatcherContract,
    ):
        self._dispatchers[id] = dispatcher

    def __getitem__(self, id: str) -> DispatcherContract:
        try:
            return self._dispatchers[id]
        except KeyError:
            raise ValueError(f"Dispatcher for id {id} not found")


class ZodchyOutboxProcessor:
    def __init__(
        self,
        dispatcher_registry: ZodchyOutboxDispatcherRegistry,
        tasks_for_status_reader: TasksForStatusReaderContract,
        tasks_status_updating_writer: TasksStatusUpdatingWriterContract,
    ):
        self._dispatcher_registry = dispatcher_registry
        self._tasks_status_updating_writer = tasks_status_updating_writer
        self._tasks_for_status_reader = tasks_for_status_reader

    async def __call__(self, query: GetOutboxTasks):
        done = []
        async for message_id in self._dispatch(
            await self._get_tasks_for_processing(query),
        ):
            done.append(message_id)

        if done:
            await self._register(done)

    async def _get_tasks_for_processing(
        self,
        query: GetOutboxTasks,
    ) -> list[OutboxTaskForProcessingReceived]:
        return [
            task
            async for task in tasks_ready_for_processing_reader(
                query,
                self._tasks_for_status_reader,
            )
        ]

    async def _dispatch(
        self,
        tasks: list[OutboxTaskForProcessingReceived],
    ) -> typing.AsyncGenerator[MessageId, None]:
        for task in tasks:
            dispatcher = self._dispatcher_registry[task.dispatcher_id]
            if await dispatcher.dispatch(
                CommunicationMessage(
                    id=task.message["id"],
                    routing_key=task.message["name"],
                    body=task.message["body"],
                    headers=task.message["headers"],
                )
            ):
                yield task.message["id"]

    async def _register(
        self,
        ids: list[TaskId],
    ):
        await self._tasks_status_updating_writer(
            ids=ids,
            status=TaskStatus.PROCESSED.value,
        )



import typing
import datetime
from zodchy.codex.cqea import Query
from .types import TaskId, MessageId


class LogItem(typing.TypedDict):
    task_id: TaskId
    status: str
    payload: dict | None = None


class MessageItem(typing.TypedDict):
    id: MessageId
    name: str
    body: dict
    headers: dict | None = None


class TaskItem(typing.TypedDict):
    id: TaskId
    status: str
    scheduled_at: datetime.datetime
    dispatcher_id: str
    message: MessageItem
    settings: dict | None = None


# ----------------------------- Models -----------------------------
class MessageCreationModel(typing.TypedDict):
    id: MessageId
    name: str
    body: dict
    headers: dict | None = None


class TaskCreationModel(typing.TypedDict):
    id: TaskId
    message_id: MessageId
    status: str
    scheduled_at: datetime.datetime
    dispatcher_id: str
    settings: dict | None = None


class TaskUpdatingModel(typing.TypedDict):
    id: TaskId
    status: str
    scheduled_at: datetime.datetime | None = None


# ----------------------------- HandlersContracts -----------------------------
class MessageCreationWriterContract(typing.Protocol):
    async def __call__(self, message: MessageCreationModel) -> None: ...


class TasksCreationWriterContract(typing.Protocol):
    async def __call__(self, *tasks: TaskCreationModel) -> None: ...


class TasksUpdatingWriterContract(typing.Protocol):
    async def __call__(self, *tasks: TaskUpdatingModel) -> None: ...


class TasksStatusUpdatingWriterContract(typing.Protocol):
    async def __call__(self, ids: list[TaskId], status: str) -> None: ...


class TasksForStatusReaderContract(typing.Protocol):
    async def __call__(self, query: Query, assure_status: str) -> list[TaskItem]: ...



def upgrade(schema: str = "zodchy") -> list[str]:
    return [
        f"""
    -- Create events table
        CREATE TABLE {schema}.outbox_messages (
            id UUID DEFAULT gen_random_uuid() NOT NULL,
            name VARCHAR NOT NULL,
            body JSONB NOT NULL,
            headers JSONB,
            CONSTRAINT pk__outbox_messages PRIMARY KEY (id)
        );
        """,
        f"""
    -- Create tasks table
    CREATE TABLE {schema}.outbox_tasks (
        id UUID DEFAULT gen_random_uuid() NOT NULL,
        message_id UUID NOT NULL,
        dispatcher_id VARCHAR NOT NULL,
        status VARCHAR NOT NULL,
        scheduled_at TIMESTAMP NOT NULL,
        settings JSONB,
        CONSTRAINT pk__outbox_tasks PRIMARY KEY (id),
        CONSTRAINT fk__outbox_tasks__message_id__outbox_messages FOREIGN KEY (message_id)
            REFERENCES {schema}.outbox_messages(id)
    );
    """,
        f"""
    -- Create index on tasks.message_id
    CREATE INDEX ix__outbox_tasks__message_id ON {schema}.outbox_tasks (message_id);
    """,
        f"""
    -- Create log table
    CREATE TABLE {schema}.outbox_logs (
        id UUID DEFAULT gen_random_uuid() NOT NULL,
        task_id UUID NOT NULL,
        status VARCHAR NOT NULL,
        payload JSONB,
        CONSTRAINT pk__outbox_logs PRIMARY KEY (id),
        CONSTRAINT fk__outbox_logs__task_id__outbox_tasks FOREIGN KEY (task_id)
            REFERENCES {schema}.outbox_tasks(id)
    );
    """,
        f"""
    -- Create index on log.task_id
    CREATE INDEX ix__outbox_logs__task_id ON {schema}.outbox_logs (task_id);
        """,
    ]


def downgrade(schema: str = "zodchy") -> list[str]:
    return [
        f"""
        -- Drop index on log.task_id
        DROP INDEX IF EXISTS {schema}.ix__outbox_logs__task_id;
        """,
        f"""
        -- Drop log table
        DROP TABLE IF EXISTS {schema}.outbox_logs;
        """,
        f"""
        -- Drop index on tasks.message_id
        DROP INDEX IF EXISTS {schema}.ix__outbox_tasks__message_id;
        """,
        f"""    
        -- Drop tasks table
        DROP TABLE IF EXISTS {schema}.outbox_tasks;
        """,
        f"""
        -- Drop events table
        DROP TABLE IF EXISTS {schema}.outbox_messages;
        """,
    ]



from .. import contracts


async def message_registration_writer(
    data: contracts.messages.OutboxMessageCreated,
    message_creation_writer: contracts.storage.MessageCreationWriterContract,
) -> None:
    await message_creation_writer(
        dict(
            id=data.id,
            name=data.name,
            body=data.body,
            headers=data.headers,
        )
    )


async def task_registration_writer(
    data: contracts.messages.OutboxTaskCreated,
    tasks_creation_writer: contracts.storage.TasksCreationWriterContract,
) -> None:
    dispatcher_id = (
        type(data.dispatcher).__name__.replace("Dispatcher", "").lower()
    )
    await tasks_creation_writer(
        dict(
            id=data.id,
            message_id=data.message_id,
            status=contracts.enums.TaskStatus.SCHEDULED.value,
            scheduled_at=data.scheduled_at,
            dispatcher_id=dispatcher_id,
            settings=data.settings,
        )
    )



from . import migrations


class Migrator:
    def __init__(self, schema: str = "zodchy", version: str = "latest"):
        self._schema = schema
        self._version = version

    def upgrade(self) -> list[str]:
        return [migration for migration in migrations.v0.upgrade(self._schema)]

    def downgrade(self) -> list[str]:
        return [migration for migration in migrations.v0.downgrade(self._schema)]



import abc
import typing
import sqlalchemy

from ....contracts.storage import TaskCreationModel, TaskUpdatingModel
from ....contracts.types import TaskId
from .generic import TransactionWriter, ConnectionWriter


class LogItem(typing.TypedDict):
    task_id: TaskId
    status: str
    payload: dict | None = None


class TasksOperationWriter(TransactionWriter, abc.ABC):
    async def _save_logs(self, *logs: LogItem):
        await self._transaction.execute(
            sqlalchemy.insert(self._schema.logs).values(logs)
        )


class TasksCreationWriter(TasksOperationWriter):
    async def __call__(self, *tasks: TaskCreationModel):
        _tasks = []
        _logs = []
        for task in tasks:
            _tasks.append(
                dict(
                    id=task["id"],
                    message_id=task["message_id"],
                    dispatcher_id=task["dispatcher_id"],
                    status=task["status"],
                    scheduled_at=task["scheduled_at"],
                    settings=(
                        task["settings"]
                        if task["settings"] is not None
                        else sqlalchemy.null()
                    ),
                )
            )
            _logs.append(
                dict(
                    task_id=task["id"],
                    status=task["status"],
                    payload=dict(scheduled_at=task["scheduled_at"]),
                )
            )
        await self._transaction.execute(
            sqlalchemy.insert(self._schema.tasks).values(_tasks)
        )
        await self._save_logs(*_logs)


class TasksUpdatingWriter(TasksOperationWriter):
    async def __call__(self, *tasks: TaskUpdatingModel):
        statuses = set()
        scheduled_ats = set()
        for task in tasks:
            statuses.add(task["status"])
            if "scheduled_at" in task:
                scheduled_ats.add(task["scheduled_at"])

        data = tasks
        if len(statuses) == 1:
            data = dict(
                status=statuses.pop(),
                scheduled_at=scheduled_ats.pop() if len(scheduled_ats) > 0 else None,
            )

        if type(data) is dict:
            await self._write_one(data, [task["id"] for task in tasks])
        elif type(data) is list:
            await self._write_many(data)

    async def _write_one(self, data: dict, ids: list[TaskId]):
        sql = (
            sqlalchemy.update(self._schema.tasks)
            .where(self._schema.tasks.c.id.in_(ids))
            .values(**{k: v for k, v in data.items() if v is not None})
        )
        await self._transaction.execute(sql)

    async def _write_many(self, data: list[dict]):
        for row in data:
            _id = row.pop("id")
            sql = (
                sqlalchemy.update(self._schema.tasks)
                .where(self._schema.tasks.c.id == _id)
                .values(row)
            )
            await self._transaction.execute(sql)


class TasksStatusUpdatingWriter(ConnectionWriter):
    async def __call__(self, ids: list[TaskId], status: str):
        async with self._engine.connect() as connection:
            await connection.execute(
                sqlalchemy.update(self._schema.tasks)
                .where(self._schema.tasks.c.id.in_(ids))
                .values(status=status)
            )



import abc

from zodchy_alchemy.contracts import EngineContract
from ..config import Metadata
from ..schema import Schema


class Reader(abc.ABC):
    def __init__(self, engine: EngineContract, metadata: Metadata | None = None):
        self._schema = Schema(metadata or Metadata())
        self._engine = engine



import sqlalchemy
from zodchy.codex.cqea import Query
from zodchy_alchemy import QueryAssembler
from zodchy_alchemy.adapters.cqea import QueryAdapter

from .generic import Reader


class TasksForStatusReader(Reader):
    async def __call__(self, query: Query, assure_status: str):
        async with self._engine.connect() as connection:
            sql_query = sqlalchemy.select(
                self._schema.tasks.c.id,
                self._schema.tasks.c.scheduled_at,
                self._schema.tasks.c.dispatcher_id,
                self._schema.tasks.c.message_id,
                self._schema.tasks.c.status,
                self._schema.tasks.c.settings,
                self._schema.messages.c.name.label("message__name"),
                self._schema.messages.c.body.label("message__body"),
                self._schema.messages.c.headers.label("message__headers"),
            ).join(
                self._schema.messages,
                self._schema.tasks.c.message_id == self._schema.messages.c.id,
                isouter=True,
            )
            adapter = QueryAdapter(
                default_table=self._schema.tasks,
            )
            assembler = QueryAssembler(sql_query)
            sql = assembler(*adapter(query))
            stream = await connection.execute(sql)
            result = [
                dict(
                    id=row.id,
                    message=dict(
                        id=row.message_id,
                        name=row.message__name,
                        body=row.message__body,
                        headers=row.message__headers,
                    ),
                    dispatcher_id=row.dispatcher_id,
                    settings=row.settings,
                    status=row.status,
                    scheduled_at=row.scheduled_at,
                )
                for row in stream
            ]
            await connection.execute(
                sqlalchemy.update(self._schema.tasks)
                .where(self._schema.tasks.c.id.in_([task["id"] for task in result]))
                .values(status=assure_status)
            )
            return result



import sqlalchemy
from functools import cached_property
import sqlalchemy.dialects.postgresql
from sqlalchemy_schema_factory import factory

from .config import Metadata


class Schema:
    def __init__(self, metadata: Metadata):
        self.db_metadata = sqlalchemy.MetaData(schema=metadata.schema)

    @cached_property
    def messages(self):
        return sqlalchemy.Table(
            "outbox_messages",
            self.db_metadata,
            factory.uuid_primary_key(),
            factory.string(name="name", nullable=False),
            factory.jsonb_aware(name="body", nullable=False),
            factory.jsonb_aware(name="headers", nullable=True),
        )

    @cached_property
    def tasks(self):
        return sqlalchemy.Table(
            "outbox_tasks",
            self.db_metadata,
            factory.uuid_primary_key(),
            factory.foreign_key(to_=self.messages, name="message_id", nullable=False),
            factory.string(name="dispatcher_id", nullable=False),
            factory.string(name="status", nullable=False),
            factory.datetime(name="scheduled_at", nullable=False),
            factory.jsonb_aware(name="settings", nullable=True),
        )

    @cached_property
    def logs(self):
        return sqlalchemy.Table(
            "outbox_logs",
            self.db_metadata,
            factory.uuid_primary_key(),
            factory.foreign_key(to_=self.tasks, name="task_id", nullable=False),
            factory.string(name="status", nullable=False),
            factory.jsonb_aware(name="payload", nullable=True),
        )



import typing
import abc
from dataclasses import dataclass

import uuid

IdentifierType: typing.TypeAlias = uuid.UUID
TaskId: typing.TypeAlias = IdentifierType
MessageId: typing.TypeAlias = IdentifierType


@dataclass(frozen=True)
class Dispatcher(abc.ABC):
    pass


@dataclass(frozen=True)
class RabbitDispatcher(Dispatcher):
    pass



import collections.abc
from ..contracts.messages import GetOutboxTasks, OutboxTaskForProcessingReceived
from ..contracts.storage import (
    TasksForStatusReaderContract,
    TasksStatusUpdatingWriterContract,
)
from ..contracts.enums import TaskStatus


async def tasks_ready_for_processing_reader(
    query: GetOutboxTasks,
    handler: TasksForStatusReaderContract,
) -> collections.abc.AsyncIterable[OutboxTaskForProcessingReceived]:
    for task in await handler(
        query=query,
        assure_status=TaskStatus.IN_PROGRESS.value,
    ):
        yield OutboxTaskForProcessingReceived(
            id=task["id"],
            message=task["message"],
            scheduled_at=task["scheduled_at"],
            dispatcher_id=task["dispatcher_id"],
            settings=task["settings"],
        )


async def processed_tasks_writer(
    tasks: list[OutboxTaskForProcessingReceived],
    tasks_status_updating_writer: TasksStatusUpdatingWriterContract,
) -> None:
    await tasks_status_updating_writer(ids=[task.id for task in tasks])



import abc

from zodchy_alchemy.contracts import WriteConnectionContract
from zodchy_alchemy.contracts import EngineContract

from ..config import Metadata
from ..schema import Schema


class Writer(abc.ABC):
    def __init__(self, metadata: Metadata | None = None):
        self._schema = Schema(metadata or Metadata())


class TransactionWriter(Writer, abc.ABC):
    def __init__(
        self, transaction: WriteConnectionContract, metadata: Metadata | None = None
    ):
        super().__init__(metadata)
        self._transaction = transaction


class ConnectionWriter(Writer, abc.ABC):
    def __init__(self, engine: EngineContract, metadata: Metadata | None = None):
        super().__init__(metadata)
        self._engine = engine
